/* lcsss3.c *********************************************\
*                                                        *
*  Higher order Godunov-type scheme for 1D Conservation  *
*                 Laws of Gas Dynamics                   *
*                                                        *
*  Uses iterative weakly linearized Riemann solver WLCS  *
*                                                        *
*  author: Andrei A. Chernousov                          *
*          E-mail: <andrei99@iname.com>                  *
*          http://www.geocities.com/andrei_chernousov    *
*  Julyh 18, 2001                                        *
*                                                        *
*********************************************************/

/*

Features:

  piecewise-parabolic spatial approximation of solution within cell
  weakly linearized characteristic-based approximate Riemann solver (WLCS)
  three step Runge-Kutta advancing in time

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define K     1.4
#define K_1_2 0.2
#define K_1   0.4
#define _05K  0.7
#define _025K 0.35
#define K_K_1 3.5
#define _1_2HI 1.25

/* alias names for fluxes */
#define F1 _U1
#define F2 _U2
#define F3 _U3

float
  *U1,  *U2,  *U3,
  *U1p, *U2p, *U3p,
  *u1,  *u2,  *u3,
  *_U1, *_U2, *_U3,
  *U1_, *U2_, *U3_,

	/* x spacing and time step */
  deltaX, deltaT, deltaT_X,

  /*--- for piecewise-parabolic reconstruction ---*/
  /* finite differences of conservative variables */
  m1, m2, m3,
  w1, w2, w3,
  /* finite differences of characteristic variables */
  M1, M2, M3,
  W1, W2, W3,
  /* modified (limited) finite differences of characteristic variables */
  _M1, _M2, _M3,
  _W1, _W2, _W3,
  /* transformation matrix [S-1], etc... */
  S11, S12, S13,
  S21, S22, S23,
  S31, S32, S33,
  /* inverted transformation matrix [S-1] */
  _S11, _S12, _S13,
  _S21, _S22, _S23,
  _S31, _S32, _S33,

  /*--- for LCS solver ---*/
  P, R, U, C,	_P, _R, _U, _C, P_, R_, U_, C_, D,
  /* characteristic speeds */
  C_p, C_m, C_o,
  /* parameters */
  uA, pA, uB, pB, rC, pC,
  /* coefficients */
  aA, aB, aC,
  /* parameters on the interface */
  pD, uD, rD, cD,

  /*--- stuff ---*/
  k1, k2, k3,
  AAA, BBB, CCC, DDD;

unsigned LEN, LENN, _i, i, i_, step, numstep;
char Stage;
FILE *pF;

/*--- For monotone piecevise parabolic reconstruction ---*/

	/* constant coefficients: 4.0, 1/3, 1/6 are recommended */
#define C1 0.333333333
#define C2 0.166666667
#define BB 4.000000000

/***************
* minmod(x, y) * - macro: limiting function of Chakravarthy, Osher
***************/
/*
	- uses register variable kk defined in Reconstruction()
	- don't use expressions as arguments of this macro!
	- beware of other side effects
*/
#define minmod(x, y) ( (x*y <= 0.) ? 0.: \
  ( (kk=(y>0.)?y*BB:-y*BB, x>0.) ? ((x<kk)?x:kk) : ((-x<kk)?x:-kk) ) )

/***************
* minmod(x, y) * - function: limiting function of Chakravarthy, Osher (slow!)
***************/
/*
float ay;
float minmod(float x, float y) 
{
  if (x * y <= 0.)
    return 0.;

  ay = ((y *= BB) > 0.) ? y: -y;
  if (x > 0.)
    return (x < ay) ? x : ay;
  return (-x < ay) ? x: -ay;
} */ /* end minmod() */


/************
*  Array1D  *
************/
float *Array1D(unsigned size)
{
  float *x;

  if((x = (float *)malloc(size * sizeof(float))) == NULL) {
    fprintf(stderr, "can't allocate memory.\n");
    exit(-1);
  }
  return x;

} /* end Array1D() */


/*************
* Initialize *   initializations
*************/
void Initialize(void)
{
  char str[80];

  /* loading data from file "lcs.ini"  */
  if((pF = fopen("lcs.ini", "r")) == NULL) {
    fprintf(stderr, "can't open file \"lcs.ini\".\n");
    exit(-1);
  }
  i = 0;
  do {
    if (fgets(str, 79, pF) != NULL) {
      switch (i) {
      case 0:	LEN    = atoi(str); break;
      case 1:  deltaX = atof(str); break;
      case 2:  deltaT = atof(str); break;
      case 3:  numstep= atoi(str); break;
      default: break;
      } /* end switch() */
      i++;
    } /* end if() */
    else {
      fprintf(stderr, "error in file \"lcs.ini\".\n");
      exit(-1);
    }
  } while (i < 4);
  fclose(pF);

  LENN = LEN+1;
  deltaT_X = deltaT / deltaX;

  /* dynamic memory allocation */
  U1  = Array1D(LEN+2);  U2 = Array1D(LEN+2);  U3 = Array1D(LEN+2);
  U1p = Array1D(LEN+2);  U2p= Array1D(LEN+2);  U3p= Array1D(LEN+2);
  _U1 = Array1D(LEN+1); _U2 = Array1D(LEN+1); _U3 = Array1D(LEN+1);
  U1_ = Array1D(LEN+1); U2_ = Array1D(LEN+1); U3_ = Array1D(LEN+1);

  /* initial conditions (20:1)!! */
  U = 0.; R = 20.; P = 2000000.;
  for (i = 1; i <= LEN/2; i++) {
    U1[i] = R;
    U2[i] = R * U;
    U3[i] = P / K_1 + 0.5 * R * U * U;
  }
  R = 1.; P = 100000.;
  for ( ; i <= LEN; i++) {
    U1[i] = R;
    U2[i] = R * U;
    U3[i] = P / K_1 + 0.5 * R * U * U;
  }

  /* next stage */
  Stage = 1, u1 = U1, u2 = U2, u3 = U3;

} /* end Initialize( ) */


/***********************
* BounCondInGhostCells *
***********************/
void BounCondInGhostCells(void)
{

  /* reflective BC */
  u1[0] =   u1[1];
  u2[0] = - u2[1];
  u3[0] =   u3[1];
  u1[LENN] =   u1[LEN];
  u2[LENN] = - u2[LEN];
  u3[LENN] =   u3[LEN];

} /* end BounCondInGhostCells( ) */


/*****************
* Reconstruction *
*****************/
void Reconstruction(void)
{
register float kk;

  for (_i = 0, i = 1, i_ = 2; i < LENN; _i++, i++, i_++) {
    /* at cell centroid */
    U =  u2[i] / (R = u1[i]);
    P = (u3[i] - 0.5 * R * U * U) * K_1;
    C = sqrt(K * P / R);
    /* inverted transformation matrix [S-1] */
    AAA = 1/(C+C);  /* 1/(2C) */
    BBB = AAA / C;  /* 1/(2*C*C) */
    CCC = BBB + BBB;
    DDD = 0.5 * U * U * BBB; /* U * U / ( 4 * C * C ) */
    _S11 = -(BBB+BBB); _S12 = _S13 = BBB;
    _S21 = -U*CCC;     _S22 = _S23 = U * BBB;      _S22 += AAA;               _S23 -= AAA;
    _S31 = -(DDD+DDD); _S32 = _S33 = DDD + _1_2HI; _S32 += ( AAA = U * AAA ); _S33 -= AAA;
    /* see "matrices" file */
    /*	  _S11 = -1/(C*C);       _S12 = 1/(2*C*C);                           _S13 = 1/(2*C*C); */
    /*	  _S21 = -U/(C*C);       _S22 = U/(2*C*C)+1/(2*C);                   _S23 = U/(2*C*C)-1/(2*C); */
    /*	  _S31 = -(U*U)/(2*C*C); _S32 = U*U/(4*C*C) + U/(2*C) + 1/(2*(K-1)); _S33 = U*U/(4*C*C) - U/(2*C) + 1/(2*(K-1)); */
    /* transformation matrix [S] */
    AAA = K_1_2*U*U;
    BBB = C*U;
    CCC = -K_1*U;
    S11 = -C*C + AAA;  S12 = CCC;    /*S13 = K_1;*/
    S21 = -BBB + AAA;  S22 = CCC+C;  /*S23 = K_1;*/
    S31 =  BBB + AAA;  S32 = CCC-C;  /*S33 = K_1;*/
    /* see "matrices" file */
    /*	  S11 = -C*C + (K-1)*U*U/2;  S12 =  -(K-1)*U;  S13 = K-1; */
    /*	  S21 = -C*U + (K-1)*U*U/2;  S22 = C-(K-1)*U;  S23 = K-1; */
    /*	  S31 =  C*U + (K-1)*U*U/2;  S32 =-C-(K-1)*U;  S33 = K-1; */
    /* differencing of the conservative variables */
    /* forward */
    m1 = u1[i_] - u1[i];
    m2 = u2[i_] - u2[i];
    m3 = u3[i_] - u3[i];
    /* backward */
    w1 = u1[i] - u1[_i];
    w2 = u2[i] - u2[_i];
    w3 = u3[i] - u3[_i];
    /* finite differences of characteristic variables dW = S * dU*/
    /* forward */
    W1 = S11*w1 + S12*w2 + K_1*w3;
    W2 = S21*w1 + S22*w2 + K_1*w3;
    W3 = S31*w1 + S32*w2 + K_1*w3;
    /* backward */
    M1 = S11*m1 + S12*m2 + K_1*m3;
    M2 = S21*m1 + S22*m2 + K_1*m3;
    M3 = S31*m1 + S32*m2 + K_1*m3;
    /* modified characteristic differences _W */
    _M1 = minmod( M1, W1 );
    _M2 = minmod( M2, W2 );
    _M3 = minmod( M3, W3 );
    _W1 = minmod( W1, M1 );
    _W2 = minmod( W2, M2 );
    _W3 = minmod( W3, M3 );
    /* vector U to... */
    /* right */
    k1 = _M1 * C1 + _W1 * C2;
    k2 = _M2 * C1 + _W2 * C2;
    k3 = _M3 * C1 + _W3 * C2;
    _U1[ i] = u1[i] + _S11 * k1 + _S12 * k2 + _S13 * k3;
    _U2[ i] = u2[i] + _S21 * k1 + _S22 * k2 + _S23 * k3;
    _U3[ i] = u3[i] + _S31 * k1 + _S32 * k2 + _S33 * k3;
    /* left */
    k1 = _M1 * C2 + _W1 * C1;
    k2 = _M2 * C2 + _W2 * C1;
    k3 = _M3 * C2 + _W3 * C1;
    U1_[_i] = u1[i] - _S11 * k1 - _S12 * k2 - _S13 * k3;
    U2_[_i] = u2[i] - _S21 * k1 - _S22 * k2 - _S23 * k3;
    U3_[_i] = u3[i] - _S31 * k1 - _S32 * k2 - _S33 * k3;
  } /* end for() */

} /* end Reconstruction( ) */


/***********************
* BounCondOnInterfaces *  BC at cell boundaries at the domain boundary
***********************/
void BounCondOnInterfaces(void)
{

  /* reflective BC */
  _U1[0] =   U1_[0];
  _U2[0] = - U2_[0];
  _U3[0] =   U3_[0];
  U1_[LEN] =   _U1[LEN];
  U2_[LEN] = - _U2[LEN];
  U3_[LEN] =   _U3[LEN];

} /* end BounCondOnInterfaces( void ) */


/********
*  LCS  *   iterative linearized characteristic-based solver
********/
void Lcs(float _P,  float _R,  float _U,
	 float  P_, float  R_, float  U_,
	 float *P,  float *R,  float *U )
{
  /* speeds od sound */
  _C= sqrt(K * _P / _R);
  C_= sqrt(K * P_ / R_);

  /* initial guess for U at the boundary */
  /* *U = (_R*_U + R_*U_) / (_R + R_); */

  /* an alternative: method of Roe: */
  D  = sqrt(R_ / _R);
  *U = (_U + U_ * D) / (1. + D);

  /*== 1st ==*/

  /* characteristic speeds */
  C_p = *U + (C = 0.5 * ( _C + C_ ));
  C_m = *U - C;
  C_o = *U;

  /* point A */
  if (C_p > 0.) {
    uA = _U;
    pA = _P;
    aA = 1./(_R*_C);
  }
  else {
    uA = U_;
    pA = P_;
    aA = 1./(R_*C_);
  }

  /* point B */
  if (C_m > 0.) {
    uB = _U;
    pB = _P;
    aB = 1./(_R*_C);
  }
  else {
    uB = U_;
    pB = P_;
    aB = 1./(R_*C_);
  }

  /* point C */
  if (C_o > 0.) {
    rC = _R;
    pC = _P;
    aC = _C * _C;
  }
  else {
    rC = R_;
    pC = P_;
    aC = C_ * C_;
  }

  /* parameters on the interface */
  pD = (uA - uB + aB * pB + aA * pA) / (aA + aB);
  uD = uA - aA * (pD - pA);
  rD = rC + (pD - pC) / aC;
  cD = sqrt(K * pD / rD);

  /* parameters at cell interface */
  *P = pD;
  *U = uD;
  *R = rD;

} // end Lcs()


/***********
*  Fluxes  *
***********/
void Fluxes(void)
{
  for (i = 0; i < LENN; i++) {

    /* left state */
    _R =  _U1[i];
    _U =  _U2[i] / _R;
    _P = (_U3[i] - 0.5 * _R * _U * _U ) * K_1;
    /* right state */
    R_ =  U1_[i];
    U_ =  U2_[i] / R_;
    P_ = (U3_[i] - 0.5 * R_ * U_ * U_) * K_1;

    /* linearized characteristic-based  solver (LCS) */
    Lcs(_P, _R, _U, P_, R_, U_, &P, &R, &U);

    /* fluxes */
    F1[i] = R * U;
    F2[i] = R * U * U + P;
    F3[i] = U * (K_K_1 * P + 0.5 * R * U * U);

    /* fluxes */
    F1[i] = R * U;
    F2[i] = R * U * U + P;
    F3[i] = U * (K_K_1 * P + 0.5 * R * U * U);

  } /* end for() */

} /* end Fluxes( ) */

/**************
*  Evolution  *
**************/
void Evolution(void)
{
  switch (Stage) {

  case 1:
    for (_i = 0, i = 1; i < LENN; i++, _i++) {
      U1p[i] = U1[i] + deltaT_X * (F1[_i] - F1[i]);
      U2p[i] = U2[i] + deltaT_X * (F2[_i] - F2[i]);
      U3p[i] = U3[i] + deltaT_X * (F3[_i] - F3[i]);
    }
    /* next stage */
    Stage = 2, u1 = U1p, u2 = U2p, u3 = U3p;
    break;

  case 2:
    for (_i = 0, i = 1; i <= LEN; i++, _i++) {
      U1p[i] = 0.75 * U1[i] + 0.25 * (U1p[i] + deltaT_X * (F1[_i] - F1[i]));
      U2p[i] = 0.75 * U2[i] + 0.25 * (U2p[i] + deltaT_X * (F2[_i] - F2[i]));
      U3p[i] = 0.75 * U3[i] + 0.25 * (U3p[i] + deltaT_X * (F3[_i] - F3[i]));
    }
    /* next stage */
    Stage = 3;
    break;

  default:
    for (_i = 0, i = 1; i <= LEN; i++, _i++) {
      U1[i] = 0.33333333333 * U1[i] + 0.66666666667 * (U1p[i] + deltaT_X * (F1[_i] - F1[i]));
      U2[i] = 0.33333333333 * U2[i] + 0.66666666667 * (U2p[i] + deltaT_X * (F2[_i] - F2[i]));
      U3[i] = 0.33333333333 * U3[i] + 0.66666666667 * (U3p[i] + deltaT_X * (F3[_i] - F3[i]));
    }
    /* next stage */
    Stage = 1, u1 = U1, u2 = U2, u3 = U3;
    break;

  } /* end switch() */

} /* end Evolution() */


/***********
*  Output  *
***********/
void Output( void )
{
  /* output to "input.dat" file*/
  if ((pF = fopen("input.dat", "w")) == NULL) {
    fprintf(stderr, "can't open \"input.dat\".\n");
    exit(-1);
  }
  fprintf(pF, "%d\n%d\n", LEN + 2, 3);
  fprintf(pF, "0. 2100000. 21.0 0. \n");
  for (i = 1; i <= LEN; i++) {
    U =  U2[i] / (R = U1[i]);
    P = (U3[i] - 0.5 * R * U * U) * K_1;
    fprintf(pF,"%3.3f %6.0f %1.4f %3.1f \n", (i-0.5)*deltaX, P, R, U);
  }
  fprintf(pF, "%d 0. 0. 561.248 \n", LEN);
  fclose(pF);

} /* end Output() */


/*********
*  MAIN  *
**********/
main()
{
  /*--- initialization ---*/
  Initialize();

	/*---*/
  while (step < numstep) {

    /*--- step increment ---*/
    printf( "%d\n", ++step );

    /*--- first stage ---*/

    /*--- BC in ghost cells ---*/
    BounCondInGhostCells( );
    /*--- parameters at cell interfaces ---*/
    Reconstruction( );
    /*---*/
    BounCondOnInterfaces( );
    /*--- fluxes at cell interfaces ---*/
    Fluxes( );
    /*--- evolution ---*/
    Evolution( );

    /*--- second stage ---*/

    /*--- BC in ghost cells ---*/
    BounCondInGhostCells( );
    /*--- parameters at cell interfaces ---*/
    Reconstruction( );
    /*---*/
    BounCondOnInterfaces( );
    /*--- fluxes at cell interfaces ---*/
    Fluxes( );
    /*--- evolution ---*/
    Evolution( );

    /*--- third stage ---*/

    /*--- BC in ghost cells ---*/
    BounCondInGhostCells( );
    /*--- parameters at cell interfaces ---*/
    Reconstruction( );
    /*---*/
    BounCondOnInterfaces( );
    /*--- fluxes at cell interfaces ---*/
    Fluxes( );
    /*--- evolution ---*/
    Evolution( );

  } /* end while */

  /*--- output flowfield ---*/
  Output();

  /**/
  return 0;

}  /* end main() */

/* end of "lcsss3.c" */
